from test.components.test_accumulate import Accumulate
from test.components.test_add_value import AddValue
from test.components.test_below import Below
from test.components.test_double import Double
from test.components.test_greet import Greet
from test.components.test_merge import Merge
from test.components.test_remainder import Remainder
from test.components.test_rename import Rename
from test.components.test_repeat import Repeat
from test.components.test_subtract import Subtract
from test.components.test_sum import Sum
